"""Implement Stream Class definition."""
from typing import List

from factiva.core import UserKey, StreamResponse, const
from factiva.news.bulknews import BulkNewsQuery
from factiva.core import req

from .subscription import Subscription


class Stream:
    """Represent a Stream workflow for Factiva API.

    Parameters
    ----------
    stream_id: str
        represents a given stream by its id
        if exists there is no need to have
        a query or a given snapshot id
    snapshot_id: str
        represents a snapshot by its id
        if exists it will be used
        to create a new stream
    query: str
        represents a query
        if exists it will be used
        to create a new stream
    user_key: Stream User
        constructor will asign a stream user which has the access
        to Pubsub client, authentication headers and urls

    Auth method 1 (All required)
    user_key: str
        if the user_key is not passed
        it can be created based on the api key param
    user_stats: bool
        if the user_key is not passed
        it can be created based on the request info param

    Auth method 2 (All required)
    user_id: str
        if the user_key is not passed
        it can be created based on the user id param
    client_id: str
        if the user_key is not passed
        it can be created based on the client id param
    password: str
        if the user_key is not passed
        it can be created based on the password param

    Examples
    --------
    Creating a new Stream directly:
        >>> stream_query_test = Stream(
            user_key='abcd1234abcd1234abcd1234abcd1234',
            user_stats=True,
            snapshot_id='<snapshot-id>',
            )
        >>> print(stream_query_test.create())
        >>>                       attributes                          id      relationships                                     type
        job_status     JOB_STATE_PENDING  dj-synhub-extraction-...                NaN                                          stream
        subscriptions                NaN  dj-synhub-extraction-...          {'data': [{'id': 'dj-synhub-extraction-...         stream

    """
    stream_id = None
    user_key = None
    snapshot_id = None
    listener = None
    subscriptions = dict()

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        stream_id=None,
        snapshot_id=None,
        query='',
        user_key=None,
        user_stats=False
    ):
        """Construct Stream class."""
        if (stream_id and (snapshot_id or query)):
            raise ValueError(
                'Not allowed stream id with query or snapshot'
            )
        self.stream_id = stream_id
        self.snapshot_id = snapshot_id
        self.query = BulkNewsQuery(query)
        self.user_key = UserKey.create_user_key(user_key, user_stats)
        if not self.user_key:
            raise RuntimeError('Undefined Stream User')

        if stream_id:
            self.set_all_subscriptions()

    @property
    def stream_url(self) -> str:
        """List Stream's URL address."""
        return f'{const.API_HOST}{const.API_STREAMS_BASEPATH}'

    @property
    def all_subscriptions(self) -> List[str]:
        """List all subscriptions to a stream."""
        return [sub.__repr__() for sub in self.subscriptions.values()]

    def get_info(self) -> StreamResponse:
        """Query a stream by its id.

        Returns
        -------
        StreamResponse which contains all information
            of the current sream

        Raises
        ------
        ValueError: when stream id is undefined
        RuntimeError: when the stream does not exists
        RuntimeError: when exists an unexpected HTTP error

        """
        if not self.stream_id:
            raise const.UNDEFINED_STREAM_ID_ERROR
        uri = '{}/{}'.format(self.stream_url, self.stream_id)
        headers = {
                'user-key': self.user_key.key
            }
        response = req.api_send_request(
            method='GET',
            endpoint_url=uri,
            headers=headers
        )
        if response.status_code == 200:
            response = response.json()
            return StreamResponse(data=response['data'], links=response.get('links', None))

        raise RuntimeError(response.text)

    def delete(self) -> StreamResponse:
        """Delete a stream.

        Returns
        -------
        StreamResponse which contains all information
            of the current which is expected
            to be CANCELLED

        Raises
        ------
        ValueError: when stream id is undefined
        RuntimeError: when the stream does not exists
        RuntimeError: when exists an unexpected HTTP error

        """
        if not self.stream_id:
            raise const.UNDEFINED_STREAM_ID_ERROR

        uri = f'{self.stream_url}/{self.stream_id}'
        headers = {
                'user-key': self.user_key.key,
                'content-type': 'application/json'
            }
        response = req.api_send_request(
            method='DELETE',
            endpoint_url=uri,
            headers=headers,
        )
        if response.status_code == 200:
            response = response.json()
            return StreamResponse(data=response['data'], links=response.get('links', None))

        if response == 404:
            raise RuntimeError('The Stream does not exist')

        raise const.UNEXPECTED_HTTP_ERROR

    def create(self) -> StreamResponse:
        """Create a stream subscription.

        There are two available options:
        Create a stream using a query
        Create a stream using a snapshot id

        Returns
        -------
        StreamResponse which contains all information
            of the current sream

        Raises
        ------
        ValueError: snapshot_id and query are undefined

        """
        if not self.snapshot_id and not self.query:
            raise ValueError('Snapshot id and query not found')
        if self.snapshot_id:
            return self._create_by_snapshot_id()

        return self._create_by_query()

    def create_subscription(self) -> str:
        """Create another subscription for an existing stream.

        Returns
        -------
        String which represents
            the new subscription id

        Raises
        ------
        RuntimeError: when unable to create a subscription

        """
        try:
            new_subscription = Subscription(stream_id=self.stream_id)
            headers = {
                'user-key': self.user_key.key
            }
            new_subscription.create(
                headers=headers
            )
            new_subscription.create_listener(self.user_key)
            self.subscriptions[new_subscription.id] = new_subscription
            return new_subscription.id
        except Exception as error:
            raise RuntimeError(
                f'''
                Unexpected error happened while
                creating the subscription: {error}
                '''
            )

    def delete_subscription(self, sus_id) -> bool:
        """Delete subscription for an existing stream.

        Parameters
        ----------
        sus_id:  str
            is the representation of a given
            subscription planned to be deleted

        Returns
        -------
        boolean which represents if the delete
            was successfully done

        Raises
        ------
        ValueError: when there is invalid subscription id
        RuntimeError: when unable to delete a subscription

        """
        if sus_id not in self.subscriptions:
            raise const.INVALID_SUBSCRIPTION_ID_ERROR
        try:
            if self.subscriptions[sus_id].delete(
                headers={'user-key': self.user_key.key}
            ):
                del self.subscriptions[sus_id]
                return True
        except Exception:
            raise RuntimeError('Unable to delete subscription')
        return False

    def create_default_subscription(self, response):
        """Create the default subscriptions at initialization.

        Adds the subscriptions to subscriptions dict

        Parameters
        ----------
        response :  dict
            is used for setting every subscription
            which exists inside the stream

        """
        for subscription in response['data']['relationships']['subscriptions']['data']:
            subscription_obj = Subscription(
                id=subscription['id'],
                stream_id=self.stream_id,
                subscription_type=subscription['type'],
                )
            subscription_obj.create_listener(self.user_key)
            self.subscriptions[subscription_obj.id] = subscription_obj

    def set_all_subscriptions(self):
        """Allow a user to set all subscriptions from a stream to local storage.

        Returns
        -------
        Dataframe which contains the state about the current stream

        Raises
        ------
        ValueError: when stream id is undefined

        """
        if not self.stream_id:
            raise const.UNDEFINED_STREAM_ID_ERROR
        uri = '{}/{}'.format(self.stream_url, self.stream_id)
        headers = {
                'user-key': self.user_key.key
            }
        response = req.api_send_request(
            method='GET',
            endpoint_url=uri,
            headers=headers
        )
        if response.status_code == 200:
            response = response.json()
            self.create_default_subscription(response)
        else:
            raise const.UNEXPECTED_HTTP_ERROR

    def consume_messages(
        self,
        callback=None,
        subscription_id=None,
        maximum_messages=None,
        batch_size=None,
        ack_enabled=False,
    ):
        """Consume messages (News) from a pubsub subscription in sync.

        Parameters
        ----------
        callback :  function
            is used for processing a message
        subscription_id :  str
            is used for connecting to pubsub
        maximum_messages : int
            is used for consuming a specific
            number of messages
        batch_size : int
            the limit of the batch expected
        ack_enabled :  boolean
            is used for acknowledging a message

        Raises
        ------
        ValueError: when subscription id is invalid

        """
        if subscription_id not in self.subscriptions:
            raise const.INVALID_SUBSCRIPTION_ID_ERROR
        self.subscriptions[subscription_id].consume_messages(
            callback=callback,
            maximum_messages=maximum_messages,
            batch_size=batch_size,
            ack_enabled=ack_enabled,
            )

    def consume_async_messages(
        self,
        callback=None,
        subscription_id=None,
        ack_enabled=False
    ):
        """Consume messages (News) from a pubsub subscription in async.

        Parameters
        ----------
        callback :  function
            is used for processing a message
        subscription_id :  str
            is used for connecting to pubsub
        ack_enabled :  boolean
            is used for acknowledging a message

        Raises
        ------
        ValueError: when subscription id is invalid

        """
        if subscription_id not in self.subscriptions:
            raise const.INVALID_SUBSCRIPTION_ID_ERROR
        self.subscriptions[subscription_id].consume_async_messages(
            callback=callback,
            ack_enabled=ack_enabled,
            )

    def _create_by_snapshot_id(self) -> StreamResponse:
        """Create by snapshot id that allows a user to create a stream subscription using a snapshot id.

        Returns
        -------
        StreamResponse
            which contains all information of the current stream

        Raises
        ------
        ValueError: When query is undefined
        RuntimeError: When API request returns unexpected error

        """
        print(self.snapshot_id)
        if not self.snapshot_id:
            raise ValueError('create fails: snaphot_id undefined')

        headers = {
                'user-key': self.user_key.key,
                'content-type': 'application/json'
            }
        uri = f'{const.API_HOST}{const.API_SNAPSHOTS_BASEPATH}/{self.snapshot_id}/streams'
        response = req.api_send_request(
            method='POST',
            endpoint_url=uri,
            headers=headers,
        )
        if response.status_code == 201:
            response = response.json()
            self.stream_id = response['data']['id']
            self.create_default_subscription(response)

            return StreamResponse(data=response['data'], links=response.get('links', None))

        raise const.UNEXPECTED_HTTP_ERROR

    def _create_by_query(self) -> StreamResponse:
        """Create by query that allows a user to create a stream subscription using a query.

        Returns
        -------
        StreamResponse
            which contains all information of the current stream

        Raises
        ------
        ValueError: When query is undefined
        RuntimeError: When API request returns unexpected error

        """
        if not self.query:
            raise ValueError('Streams query undefined in Create by query')

        base_query = self.query.get_base_query()
        streams_query = {
            "data": {
                "attributes": base_query['query'],
                "type": "stream"
                }
            }

        headers = {
                'user-key': self.user_key.key,
                'content-type': 'application/json'
            }
        response = req.api_send_request(
                method='POST',
                endpoint_url=self.stream_url,
                headers=headers,
                payload=streams_query,
            )

        if response.status_code == 201:
            response = response.json()
            self.stream_id = response['data']['id']
            self.create_default_subscription(response)

            return StreamResponse(data=response['data'], links=response.get('links', None))

        raise const.UNEXPECTED_HTTP_ERROR



"""Implement Listener class."""
import json
import time

from threading import Thread

from google.api_core.exceptions import GoogleAPICallError, NotFound

from factiva.core.tools import load_environment_value
from factiva.core import const
from factiva.core import req


def default_callback(message, subscription_id):
    """Call to default callback function."""
    print('Subscription ID: {}: Message: {}'.format(subscription_id, message))


class Listener:
    r"""Class that represents a Listener for Google Pubsub.

    Parameters
    ----------
    stream_user : Stream User
        constructor will asign a stream user which has the access
        to the proper url and headers which are going to be used for:

        - Checking the exceeded documents
        - Consuming messages (articles) in sync
        - Consuming messages (articles) in async

    subscription_id :  str
        is used by Pubsub
        to consume messages in async/sync

    Examples
    --------
    Creating a new Listener directly:
        >>> listener = Listener(
                stream_user=StreamUser(
                    user_key='****************************1234',
                    request_info=False,
                    user_id='******-svcaccount@dowjones.com',
                    client_id='****************************5678',
                    password='*******'
                )
            )
        >>> def callback(message, subscription_id):
        >>>     print('Subscription ID: {}: Message: {}'.format(
                    subscription_id, message
                ))
        >>> print(listener.listen(
                callback,
                subscription_id='<subscription-id>',
                maximum_messages=10
                ))
        Received news message with ID: DJDN******************
        Subscription ID: dj-synhub-stream-********************-
        km******-filtered-******:
        Message: {'an': 'DJDN0*********************',
        'document_type': 'article', 'action': 'rep',
        'source_code': 'DJDN', 'source_name': 'Dow Jones ---- -----',
        'publication_date': '2021-05-20T08:00:10.255Z',
        'publication_datetime': '2021-05-20T08:00:10.255Z',
        'modification_date': '2021-05-20T08:04:56.175Z',
        'modification_datetime': '2021-05-20T08:02:54.000Z',
        'ingestion_datetime': '2021-05-20T08:00:13.000Z',
        'title': "----- ------ ------ ------ 2020",
        'snippet': '', 'body': "\nOn Thursday -- --- ----,
        --- Plc. announced its ----- ---- --- \n
        ......
        }

    """

    _check_exceeds_thread = None
    FIRST_OBJECT = 0

    def __init__(self, subscription_id=None, user_key=None):
        """Instantiate listener class constructor."""
        if not subscription_id:
            try:
                subscription_id = load_environment_value('FACTIVA_STREAM_SUBSCRIPTION_ID')
            except Exception:
                raise const.UNDEFINED_SUBSCRIPTION_ERROR

        if not user_key:
            raise ValueError('Undefined user_key')

        self.user_key = user_key
        self.subscription_id = subscription_id
        self.is_consuming = True
        self.limit_msg = None

    @property
    def stream_id_uri(self):
        """Property for retrieving the stream id uri."""
        host = self.user_key.get_uri_context()
        stream_id = '-'.join(self.subscription_id.split("-")[:-2])
        return f'{host}/streams/{stream_id}'

    def _check_account_status(self):
        """Check the account status for max allowed extracts done.

        Raises
        ------
        RuntimeError: When HTTP API Response is unexpected

        """
        # TODO: Implement using UserKey.get_stats()
        host = self.user_key.get_uri_context()
        headers = self.user_key.get_authentication_headers()
        limits_uri = f'{host}/accounts/{self.user_key.user_key}'
        limit_response = req.api_send_request(
            method='GET',
            endpoint_url=limits_uri,
            headers=headers
            )
        if limit_response.status_code == 200:
            limit_response = limit_response.json()
            self.limit_msg = limit_response['data']['attributes']['max_allowed_extracts']
        else:
            raise RuntimeError(
                '''
                Unexpected HTTP Response from API
                while checking for limits
                '''
                )

    def _check_stream_status(self):
        """Check the stream status.

        if it has reached an
        exceeded status at some point

        Raises
        ------
        RuntimeError: When HTTP API Response is unexpected

        """
        headers = self.user_key.get_authentication_headers()
        response = req.api_send_request(
            method='GET',
            endpoint_url=self.stream_id_uri,
            headers=headers
            )
        if response.status_code == 200:
            response = response.json()
            job_status = response['data']['attributes']['job_status']
            if job_status == const.DOC_COUNT_EXCEEDED:
                self._check_account_status()
        else:
            raise RuntimeError('HTTP API Response unexpected')

    def _check_exceeded(self):
        """Check exceeded time function.

        checks if the documents have been exceeded
        (max allowed extractions exceeded)
        """
        while self.is_consuming:
            print('Checking if extractions limit is reached')
            self._check_stream_status()
            time.sleep(const.CHECK_EXCEEDED_WAIT_SPACING)
        if not self.limit_msg:
            print('Job finished')
        else:
            print(
                '''
                OOPS! Looks like you\'ve exceeded the maximum number of
                documents received for your account ({}). As such, no
                new documents will be added to your stream\'s queue.
                However, you won\'t lose access to any documents that
                have already been added to the queue. These will continue
                to be streamed to you. Contact your account administrator
                with any questions or to upgrade your account limits.
                '''.format(self.limit_msg)
            )

    def check_exceeded_thread(self):
        """Check exceeded thread function.

        creates threads for checking
        if the doc count has been exceeded

        """
        self._check_exceeds_thread = Thread(target=self._check_exceeded)
        self._check_exceeds_thread.start()

    # pylint: disable=too-many-arguments
    def _pull_pubsub_messages(
        self,
        pubsub_client,
        pubsub_request,
        subscription_path,
        callback,
        ack_enabled
    ):
        """Pull messages from pubsub.

        if any message exist for
        the given subscription id

        Parameters
        ----------
        pubsub_client:  Google Pubsub client
            is used for consuming
            pubsub messages
        pubsub_request: object
            which represents a request for
            google pubsub
        subscription_path: str
            represents the path inside pubsub
            for the current subscription
        callback: function
            function executed
            by message received
        ack_enabled: bool
            a listener can consume a message
            again if ack_enabled is false,
            otherwise it won't

        """
        pubsub_messages = pubsub_client.pull(request=pubsub_request)
        if pubsub_messages and pubsub_messages.received_messages:
            for message in pubsub_messages.received_messages:
                pubsub_message = json.loads(message.message.data)
                print("Received news message with ID: {}".format(
                    pubsub_message['data'][self.FIRST_OBJECT]['id'])
                )
                news_message = pubsub_message['data'][self.FIRST_OBJECT]['attributes']
                callback_result = callback(
                    news_message,
                    self.subscription_id
                    )
                if ack_enabled:
                    pubsub_client.acknowledge(
                        subscription=subscription_path,
                        ack_ids=[message.ack_id]
                        )
                self.messages_count += 1
                if not callback_result:
                    return

                print(f'Callback returns: {callback_result}')

    def listen(
        self,
        callback=default_callback,
        maximum_messages=None,
        batch_size=10,
        ack_enabled=False
    ):
        """Listen function.

        listens the current messages (News)
        from a pubsub subscription in sync

        Parameters
        ----------
        callback :  function
            is used for processing a message
        maximum_messages: int
            is used for consuming a specific
            number of messages
        batch_size: int
            the limit of the batch expected
        ack_enabled: boolean
            flag for acknowledge
            a message

        Raises
        ------
        ValueError:
            When maximum_messages is undefined
        GoogleAPICallError:
            When there is no valid instance to pull from
            When something unexpected happened with Pubsub client

        """
        if not maximum_messages:
            raise ValueError('undefined maximum messages to proceed')

        pubsub_client = self.user_key.get_client_subscription()
        self.check_exceeded_thread()

        streaming_credentials = self.user_key.fetch_credentials()
        subscription_path = pubsub_client.subscription_path(
            streaming_credentials['project_id'],
            self.subscription_id
            )
        print(
            '''
            Listeners for subscriptions have been set up
            and await message arrival.
            '''
            )

        self.messages_count = 0
        pubsub_request = {
            "subscription": subscription_path,
            "max_messages": batch_size,
            "return_immediately": True
            }
        while (maximum_messages is None) or (self.messages_count < maximum_messages):
            try:
                if maximum_messages is not None:
                    maximum_messages = min(
                        batch_size,
                        maximum_messages - self.messages_count
                        )
                    self._pull_pubsub_messages(
                        pubsub_client,
                        pubsub_request,
                        subscription_path,
                        callback,
                        ack_enabled
                    )
            except GoogleAPICallError as google_error:
                if isinstance(google_error, NotFound):
                    raise google_error
                print(
                    '''
                    Encountered a problem while trying to pull a message
                    from a stream. Error is as follows: {}
                    '''.format(str(google_error))
                    )
                print(
                    '''
                    Due to the previous error, system will pause 10 seconds.
                    System will then attempt to pull the message from
                    the stream again.
                    '''
                    )
                time.sleep(const.PUBSUB_MESSAGES_WAIT_SPACING)
                pubsub_client = self.user_key.get_client_subscription()

        self.is_consuming = False

    def listen_async(self, callback=default_callback, ack_enabled=False):
        """Listen async function.

        listens the current messages (News)
        from a pubsub subscription in async

        Parameters
        ----------
        callback:  function
            is used for processing a message
        ack_enabled: boolean
            flag for acknowledge
            a message

        """
        def ack_message_and_callback(message):
            pubsub_message = json.loads(message.data)
            print("Received news message with ID: {}".format(
                pubsub_message['data'][self.FIRST_OBJECT]['id']
                )
            )
            news_message = pubsub_message['data'][self.FIRST_OBJECT]['attributes']
            callback(news_message, self.subscription_id)
            if ack_enabled:
                message.ack()

        pubsub_client = self.user_key.get_client_subscription()
        self.check_exceeded_thread()

        streaming_credentials = self.user_key.fetch_credentials()
        subscription_path = pubsub_client.subscription_path(
            streaming_credentials['project_id'],
            self.subscription_id
            )
        pubsub_client.subscribe(
            subscription_path,
            callback=ack_message_and_callback
            )
        print(
            '''
            Listeners for subscriptions have been set up
            and await message arrival.
            '''
            )



"""Represent a Factiva Snapshot Class."""
from factiva.news.bulknews import BulkNewsBase
from .query import SnapshotQuery
from .jobs import ExplainJob, ExtractionJob, AnalyticsJob, UpdateJob


class Snapshot(BulkNewsBase):
    """Represent a Factiva Snapshot Class.

    Class that represents a Factiva Snapshot.

    Parameters
    ----------
    user_key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    user_stats : boolean, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.
    query : str or SnapshotQuery, optional
        Query used to run any of the Snapshot-related operations. If a str is
        provided, a simple query with a `where` clause is created. If other
        query fields are required, either provide the SnapshotQuery object at
        creation, or set the appropriate object values after creation. This
        parameter is not compatible with snapshot_id.
    snapshot_id: str, optional
        String containing the 10-character long Snapshot ID. This parameter is
        not compatible with query.

    See Also
    --------
    Stream: Class that represents the continuous Factiva News document stream.

    Examples
    --------
    Creating a new Snapshot with an key string and a Where statement. Then, running a full Explain process.
        >>> from factiva.news.snapshot import Snapshot
        >>> my_key = "abcd1234abcd1234abcd1234abcd1234"
        >>> my_query = "publication_datetime >= '2020-01-01 00:00:00' AND LOWER(language_code) = 'en'"
        >>> my_snapshot = Snapshot(user_key=my_key, query=my_query)
        >>> my_snapshot.process_explain()
        106535

    Creating a new Snapshot from an UserKey and a SnapshotQuery instances. Then, running a full Analytics process.
        >>> my_user = UserKey()
        >>> my_query = SnapshotQuery("publication_datetime >= '2020-01-01 00:00:00' AND LOWER(language_code) = 'en'")
        >>> my_query.frequency = 'YEAR'
        >>> my_query.group_by_source_code = True
        >>> my_query.top = 20
        >>> my_snapshot = Snapshot(user_key=my_user, query=my_query)
        >>> analytics_df = my_snapshot.process_analytics()
        >>> analytics_df.head()
              count  publication_datetime  source_code
            0	20921	1995	NGCIOS
            1	20371	1995	LATAM
            2	18303	1995	REUTES
            3	10593	1995	EXPNSI
            4	4212	1995	MUNDO

    """
    query = None
    folder_path = ''
    file_format = ''
    file_list = []
    news_data = None
    last_explain_job = None
    last_analytics_job = None
    last_extraction_job = None
    last_update_job = None

    def __init__(
        self,
        user_key=None,
        user_stats=False,
        query=None,
        snapshot_id=None
    ):
        """Instantiate class."""
        super().__init__(user_key=user_key, user_stats=user_stats)

        self.last_explain_job = ExplainJob(user_key=self.user_key)
        self.last_analytics_job = AnalyticsJob(user_key=self.user_key)

        if query and snapshot_id:
            raise Exception("The query and snapshot_id parameters cannot be set simultaneously")

        if query:
            if isinstance(query, SnapshotQuery):
                self.query = query
            elif isinstance(query, str):
                self.query = SnapshotQuery(query)
            else:
                raise ValueError("Unexpected value for the query-where clause")
            self.last_extraction_job = ExtractionJob(user_key=self.user_key)

        if snapshot_id:
            self.query = SnapshotQuery('')
            self.last_extraction_job = ExtractionJob(snapshot_id=snapshot_id, user_key=self.user_key)
            self.get_extraction_job_results()

    def submit_explain_job(self):
        """Submit an Explain job to the Factiva Snapshots API.

        Submits an Explain job to the Factiva Snapshots API, using the
        assigned user in the `user_key`, and SnapshotQuery in the `query`
        properties.

        Returns
        -------
        Boolean : True if the submission was successful. An Exception otherwise.

        """
        return self.last_explain_job.submit_job(self.query.get_explain_query())

    def get_explain_job_results(self):
        """Obtain the Explain job results from the Factiva Snapshots API.

        Obtains the Explain job results from the Factiva Snapshots API. Results
        are stored in the `last_explain_job` class property.

        Returns
        -------
        Boolean : True if the data was retrieved successfully. An Exception
            otherwise.

        """
        return self.last_explain_job.get_job_results()

    def process_explain(self):
        """Submit an Explain job to the Factiva Snapshots API.

        Submits an Explain job to the Factiva Snapshots API, using the same
        parameters used by `submit_explain_job`. Then, monitors the job until
        its status change to `JOB_STATE_DONE`. Finally, retrieves and stores
        the results in the property `last_explain_job`.

        Returns
        -------
        Boolean : True if the explain processing was successful. An Exception
            otherwise.

        Examples
        --------
        Process explain job from snapshot
            >>> query_clause = "publication_datetime >= '2018-01-01 00:00:00' AND publication_datetime <= '2018-01-02 00:00:00' AND LOWER(language_code) = 'en'"
            >>> my_snapshot = Snapshot(user_key='abcd1234abcd1234abcd1234abcd1234', query=query_clause)
            >>> try:
            ... 	my_snapshot.process_explain()
            >>> except RuntimeError:
            ...		print('There was an error with the API call')
            >>>
            >>> print(my_snapshot.last_explain_job.document_volume)
            450483

        """
        return self.last_explain_job.process_job(self.query.get_explain_query())

    def submit_analytics_job(self):
        """Submit an Analytics job to the Factiva Snapshots API.

        Submits an Analytics job to the Factiva Snapshots API, using the
        assigned user in the `user_key`, and SnapshotQuery in the `query`
        properties.

        Returns
        -------
        Boolean : True if the submission was successful. An Exception otherwise.

        """
        return self.last_analytics_job.submit_job(self.query.get_analytics_query())

    def get_analytics_job_results(self):
        """Obtain the Analytics job results from the Factiva Snapshots API.

        Obtains the Analytics job results from the Factiva Snapshots API.
        Results are stored in the `last_analytics_job` class property.

        Returns
        -------
        Boolean : True if the data was retrieved successfully. An Exception
            otherwise.

        """
        return self.last_analytics_job.get_job_results()

    def process_analytics(self):
        """Submit an Analytics job to the Factiva Snapshots API.

        Submits an Analytics job to the Factiva Snapshots API, using the same
        parameters used by `submit_analytics_job`. Then, monitors the job until
        its status change to `JOB_STATE_DONE`. Finally, retrieves and stores
        the results in the property `last_analytics_job`.

        Returns
        -------
        Boolean : True if the analytics processing was successful. An Exception
            otherwise.

        Examples
        --------
        Process analytics job
            >>> query_clause = "publication_datetime >= '2018-01-01 00:00:00' AND publication_datetime <= '2018-01-02 00:00:00' AND LOWER(language_code) = 'en'"
            >>> my_snapshot = Snapshot(user_key='abcd1234abcd1234abcd1234abcd1234', query=query_clause)
            >>> my_snapshot.process_analytics()
            >>> print(my_snapshot.last_analytics_job.data)
                publication_datetime   count
            0               2018-01  950516
            1               2018-02  929795
            2               2018-03  998663
            3               2018-04  935845
            4               2018-05  894903
            5               2018-06  876938
            6               2018-07  867509
            7               2018-08  793283
            8               2018-09  858963
            9               2018-10  957739
            10              2018-11  917355
            11              2018-12   38401

        """
        return self.last_analytics_job.process_job(self.query.get_analytics_query())

    def submit_extraction_job(self):
        """Submit an Extraction job to the Factiva Snapshots API.

        Submits an Extraction job to the Factiva Snapshots API, using the
        assigned user in the `user_key`, and SnapshotQuery in the `query`
        properties.

        Returns
        -------
        Boolean : True if the submission was successful. An Exception otherwise.

        """
        is_successful = self.last_extraction_job.submit_job(self.query.get_extraction_query())

        if is_successful:
            self.snapshot_id = self.last_extraction_job.job_id
        return is_successful

    def get_extraction_job_results(self):
        """Obtain the Extraction job results from the Factiva Snapshots API.

        Obtains the Extraction job results from the Factiva Snapshots API.
        Results are stored in the `last_extraction_job` class property.

        Returns
        -------
        Boolean : True if the data was retrieved successfully. An Exception
            otherwise.

        """
        return self.last_extraction_job.get_job_results()

    def download_extraction_files(self, download_path=None):
        """Download the list of files listed in the Snapshot.last_extraction_job.files.

        Downloads the list of files listed in the Snapshot.last_extraction_job.files
        property, and stores them in a folder indicated by `download_path`. If no
        `download_path` is provided, then files are stored in a folder with the
        same name as the snapshot ID.

        Parameters
        ----------
        download_path: str, optional
            String containing the file path on where to store the files. If not
            provided, files are stored in a folder with the same name as the
            update ID.

        Returns
        -------
        Boolean : True if the files were correctly downloaded, False otherwise.

        """
        return self.last_extraction_job.download_job_files(download_path)

    def process_extraction(self, download_path=None):
        """Submit an Extraction job to the Factiva Snapshots API.

        Submits an Extraction job to the Factiva Snapshots API, using the same
        parameters used by `submit_extraction_job`. Then, monitors the job until
        its status change to `JOB_STATE_DONE`. The final status is retrieved
        and stored in the property `last_extraction_job`, which among other
        properties, contains the list of files to download. The process then
        downloads all files to the specified `download_path`. If no download path
        is provided, files are stored in a folder named equal to the `snapshot_id`
        property. The process ends after all files are downloaded.

        Because the whole processing takes places in a single call, it's
        expected that the execution of this operation takes several
        minutes, or even hours.

        Parameters
        ----------
        download_path: str, optional
            String containing the file path on where to store the files. If not
            provided, files are stored in a folder with the same name as the
            update ID.

        Returns
        -------
        Boolean : True if the extraction processing was successful. An Exception
            otherwise.

        Examples
        --------
        Process extraction job.
            >>> query_clause = "publication_datetime >= '2018-01-01 00:00:00' AND publication_datetime <= '2018-01-02 00:00:00' AND LOWER(language_code) = 'en'"
            >>> my_snapshot = Snapshot(user_key='abcd1234abcd1234abcd1234abcd1234', query=query_clause)
            >>> my_snapshot.process_extraction(path='../downloads/data')

        """
        return self.last_extraction_job.process_job(self.query.get_extraction_query(), download_path)

    def submit_update_job(self, update_type):
        """Submit an Update Job to the Factiva Snapshots API.

        Submits an Update Job to the Factiva Snapshots API, using the
        assigned user in the `user_key` and `snapshot_id` asigned to
        the instance and the `update_type` passed as parameter. Assigns
        the submitted job to the `last_update_job` property.

        Parameters
        ----------
        update_type: str
            String containing the update type to submit a job.
            Could be 'additions', 'replacements' or 'deletes'.

        Returns
        -------
        Boolean : True if the submission was successful. An Exception otherwise.

        """
        self.last_update_job = UpdateJob(update_type=update_type, snapshot_id=self.last_extraction_job.job_id)
        return self.last_update_job.submit_job()

    def get_update_job_results(self):
        """Obtain of the Update Job results from the Factiva Snapshots API.

        Obtains the Update Job results from the Factiva Snapshots API.
        Results are stored in the `last_update_job` class property.

        Raises
        ------
        - RuntimeError when an update job has not beed submitted.

        Returns
        -------
        Boolean : True if the data was retrieved successfully. An Exception
            otherwise.

        """
        if self.last_update_job is None:
            raise RuntimeError('Update job has not been set')
        return self.last_update_job.get_job_results()

    def download_update_files(self, download_path=None):
        """Download the list of files listed in the Snapshot.last_update_job.files property.

        Download the list of files listed in the Snapshot.last_update_job.files
        property, and stores them in a folder indicated by `download_path`. If no
        `download_path` is provided, then files are stored in a folder with the
        same name as the update ID.

        Parameters
        ----------
        download_path: str, optional
            String containing the file path on where to store the files. If not
            provided, files are stored in a folder with the same name as the
            update ID.

        Raises
        ------
        - RuntimeError when an update job has not been submitted.

        Returns
        -------
        Boolean : True if the files were correctly downloaded, False otherwise.

        """
        if self.last_update_job is None:
            raise RuntimeError('Update job has not been set')
        return self.last_update_job.download_job_files(download_path)

    def process_update(self, update_type, download_path=None):
        """Submit an Update job to the Factiva Snapshots API.

        Submits an Update job to the Factiva Snapshots API using the same
        parameters used by `submit_update_job`. Then, monitors the job until
        its status change to `JOB_STATE_DONE`. The final status is retrieved
        and stored in the property `last_update_job`, which among other
        properties, contains the list of files to download. The process then
        downloads all files to the specified `download_path`. If no download path
        is provided, files are stored in a folder named equal to the
        `last_update_job.job_id` property.

        Because the whole processing takes places in a single call, it's
        expected that the execution of this operation takes several
        minutes, or even hours.

        Parameters
        ----------
        update_type: str
            String containing the update type to submit a job.
            Could be 'additions', 'replacements' or 'deletes'.
        download_path: str, optional
            String containing the file path on where to store the files. If not
            provided, files are stored in a folder with the same name as the
            update ID.

        Returns
        -------
        Boolean : True if the update processing was successful. An Exception
            otherwise.

        Examples
        --------
        Process update job with type 'additions'
            >>> previous_snapshot = Snapshot(user_key=my_user, snapshot_id='sdjjekl93j')
            >>> previous_snapshot.process_update('additions', download_path=f'./{previous_snapshot.snapshot_id}/additions/')

        """
        self.last_update_job = UpdateJob(update_type=update_type, snapshot_id=self.last_extraction_job.job_id)
        return self.last_update_job.process_job(path=download_path)

    def __repr__(self):
        """Create string representation for Snapshot Class."""
        return self.__str__()

    def __str__(self, detailed=True, prefix='  |-', root_prefix=''):
        """Create string representation for Snapshot Class."""
        pprop = self.__dict__.copy()
        child_prefix = '  |    |-'
        ret_val = str(self.__class__) + '\n'

        ret_val += f'{prefix}user_key: '
        ret_val += self.user_key.__str__()
        del pprop['user_key']
        ret_val += '\n'

        ret_val += f'{prefix}query: '
        ret_val += self.query.__str__(detailed=False, prefix=child_prefix)
        del pprop['query']
        ret_val += '\n'

        ret_val += f'{prefix}last_explain_job: '
        ret_val += self.last_explain_job.__str__(detailed=False, prefix=child_prefix)
        del pprop['last_explain_job']
        ret_val += '\n'

        ret_val += f'{prefix}last_analytics_job: '
        ret_val += self.last_analytics_job.__str__(detailed=False, prefix=child_prefix)
        del pprop['last_analytics_job']
        ret_val += '\n'

        ret_val += f'{prefix}last_extraction_job: '
        ret_val += self.last_extraction_job.__str__(detailed=False, prefix=child_prefix)
        del pprop['last_extraction_job']
        ret_val += '\n'

        ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        return ret_val




"""
    Init file for news module.
"""
__all__ = ['snapshot', 'stream', 'bulknews', 'taxonomy',
    'Snapshot', 'SnapshotQuery', 'SnapshotFiles', 'Stream',
    'Listener', 'Subscription', 'Taxonomy']

from .__version__ import __version__
from .snapshot import (Snapshot, SnapshotQuery, SnapshotFiles)
from .stream import (Stream, Listener, Subscription)
from .taxonomy.taxonomy import Taxonomy

version = __version__



"""
    Describe modules included for Stream Class definition.
"""
__all__ = ['Stream', 'Listener', 'Subscription', 'stream', 'listener', 'subscription']

from .stream import Stream
from .listener import Listener
from .subscription import Subscription



"""Implement Subscription class."""
from factiva.core import req
from factiva.core import const
from factiva.core import StreamUser
from factiva.core.tools import load_environment_value

from .listener import Listener


class Subscription:
    """Represent a Subscription inside a stream.

    Class that represents a Subscription inside a stream.
    There a two possible operations for a Subscription:
    - Create new one based on an existing Stream
    - Delete an existing subscription from a Stream

    Parameters
    ----------
    url: str
        url used to create/delete a subscription
    stream_id: str
        represents a given stream by its id

    Raises
    ------
    ValueError: when a stream_id is undefined

    Examples
    --------
    Creating a new Subscription directly:
        >>> subscription = Subscription(<stream_id>)
        >>> created_subs = subscription.create(
                headers={'authorization': 'user-key'}
            )
        >>> print(created_subs)
        >>> { "id": "dj-synhub-extraction-*HH**", "type": "subscription" }

    """

    SUBSCRIPTION_IDX = 0
    id = None
    stream_id = None
    subscription_type = None
    listener = None

    def __init__(self, stream_id=None, id=None, subscription_type=None):
        """Instantiate listener class constructor."""
        if not stream_id:
            try:
                stream_id = load_environment_value('FACTIVA_STREAM_SUBSCRIPTION_ID')
            except Exception:
                raise const.UNDEFINED_STREAM_ID_ERROR

        self.url = f'{const.API_HOST}{const.API_STREAMS_BASEPATH}'
        self.stream_id = stream_id
        # pylint: disable=invalid-name
        self.id = id
        self.subscription_type = subscription_type

    def __repr__(self):
        """Create string representation for Subscription Class."""
        return f'Subscription(id={self.id}, type={self.subscription_type})'

    def create_listener(self, user):
        """Create a listener in a separate step.

        Create listener allows to create a listener
        in a separate step for avoiding undefined
        subscription id

        Parameters
        ----------
        user: StreamUser
            user which possess access to any
            credentials/client needed
            for listener

        Raises
        ------
        RuntimeError: when user is not a StreamUser

        """
        if not isinstance(user, StreamUser):
            raise RuntimeError('user is not a StreamUser instance')

        self.listener = Listener(
            subscription_id=self.id,
            stream_user=user
            )

    def create(self, headers=None):
        """Create a subscription for a given stream instance.

        Create subscription allows a user to create
        another subscription to a given stream

        Parameters
        ----------
        headers: dict
            which contains the token/acces key for authorization

        Returns
        -------
        Data which contains:
        subscription's id and type created

        Raises
        ------
        ValueError: when a stream_id is undefined
        RuntimeError: when Unexpected API response happens

        """
        if not self.stream_id:
            raise ValueError(
                '''
                stream_id is not defined,
                it must be defined for creating a subscription
                '''
            )

        uri = '{}/{}/subscriptions'.format(self.url, self.stream_id)
        response = req.api_send_request(
            method='POST',
            endpoint_url=uri,
            headers=headers
            )
        if response.status_code == 201:
            response = response.json()
            data = response['data']
            self.id = data[self.SUBSCRIPTION_IDX]['id']
            self.subscription_type = data[self.SUBSCRIPTION_IDX]['type']
            return data

        raise RuntimeError('Unexpected API response')

    def delete(self, headers=None) -> bool:
        """Delete subscription for a given stream.

        Delete subscription allows a user to delete
        a subscription to a given stream

        Parameters
        ----------
        headers: dict
            which contains the token/acces key for authorization

        Returns
        -------
        bool value which shows if the subscription is complete deleted

        Raises
        ------
        RuntimeError: when Unexpected API response happens

        """
        uri = '{}/{}/subscriptions/{}'.format(
            self.url,
            self.stream_id,
            self.id
            )
        response = req.api_send_request(
            method='DELETE',
            endpoint_url=uri,
            headers=headers
            )
        if response.status_code == 200:
            return True

        raise RuntimeError('Unexpected API response')

    def consume_messages(
        self,
        callback=None,
        maximum_messages=None,
        batch_size=None,
        ack_enabled=False,
    ):
        """Consume messages from a pubsub subscription in sync.

        Consume messages is a listener function
        which consumes the current messages (News)
        from a pubsub subscription in sync

        Parameters
        ----------
        callback :  function
            is used for processing a message
        maximum_messages: int
            is used for consuming a specific
            number of messages
        batch_size: int
            the limit of the batch expected
        ack_enabled :  boolean
            is used for acknowledging a message

        Raises
        ------
        RuntimeError: when listener is not yet init

        """
        if not self.listener:
            raise RuntimeError('uninitialized listener')
        if batch_size:
            self.listener.listen(
                callback=callback,
                maximum_messages=maximum_messages,
                batch_size=batch_size,
                ack_enabled=ack_enabled,
                )
        else:
            self.listener.listen(
                callback=callback,
                maximum_messages=maximum_messages,
                ack_enabled=ack_enabled,
                )

    def consume_async_messages(self, callback=None, ack_enabled=False):
        """Consume async messages is a listener function.

        Consume async messages is a listener function
        which consumes the current messages (News)
        from a pub-sub subscription in async

        Parameters
        ----------
        callback :  function
            is used for processing a message
        ack_enabled :  boolean
            is used for acknowledging a message

        Raises
        ------
        RuntimeError: when listener is not yet init

        """
        if not self.listener:
            raise RuntimeError('uninitialized listener')
        self.listener.listen_async(
            callback=callback,
            ack_enabled=ack_enabled,
        )



import os
import pandas as pd
import fastavro


class SnapshotFiles(object):

    # TODO: This list doesn't consider company identifiers like ISIN or CUSIP
    ARTICLES_STAT_FIELDS = ['an', 'company_codes', 'company_codes_about',
                            'company_codes_occur', 'industry_codes', 'ingestion_datetime',
                            'language_code', 'modification_datetime',
                            'publication_datetime',
                            'publisher_name', 'region_codes', 'region_of_origin',
                            'source_code', 'source_name',
                            'subject_codes', 'title', 'word_count']

    # publication_date and modification_date are deprecated fields
    ARTICLE_DELETE_FIELDS = ['art', 'credit', 'document_type', 'publication_date', 'modfication_date']

    def read_file(self, filepath, only_stats=False, merge_body=False) -> pd.DataFrame:
        """Reads a single Dow Jones snapshot datafile
        Parameters
        ----------
        filepath : str
            Relative or absolute file path
        only_stats : bool, optional
            Specifies if only file metadata is loaded (True), or if the full article content is loaded (False). On average,
            only_stats loads about 1/10 and is recommended for quick metadata-based analysis. (Default is False)
        merge_body : bool, optional
            Specifies if the body field should be merged with the snippet and this last column being dropped.
            (default is False)
        Returns
        -------
        pandas.DataFrame
            A single Pandas Dataframe with the file content
        """
        with open(filepath, "rb") as fp:
            reader = fastavro.reader(fp)
            records = [r for r in reader]
            r_df = pd.DataFrame.from_records(records)

        if only_stats is True:
            r_df = r_df[self.ARTICLES_STAT_FIELDS]

        if (only_stats is False) & (merge_body is True):
            r_df['body'] = r_df['snippet'] + '\n\n' + r_df['body']
            r_df.drop('snippet', axis=1, inplace=True)

        r_df['body'] = r_df[['body']].apply(lambda x: '{}'.format(x[0]), axis=1)

        for d_field in self.ARTICLE_DELETE_FIELDS:
            if d_field in r_df.columns:
                r_df.drop(d_field, axis=1, inplace=True)

        r_df['publication_date'] = r_df['publication_date'].astype('datetime64[ms]')
        r_df['publication_datetime'] = r_df['publication_datetime'].astype('datetime64[ms]')
        r_df['modification_date'] = r_df['modification_date'].astype('datetime64[ms]')
        r_df['modification_datetime'] = r_df['modification_datetime'].astype('datetime64[ms]')
        r_df['ingestion_datetime'] = r_df['ingestion_datetime'].astype('datetime64[ms]')
        return r_df

    def read_folder(self, folderpath, file_format='AVRO', only_stats=False, merge_body=False) -> pd.DataFrame:
        """Scans a folder and reads the content of all files matching the format (file_format)
        Parameters
        ----------
        folderpath : str
            Relative or absolute folder path
        file_format : str, optional
            Supported file format. Current options are AVRO, JSON or CSV. (default is AVRO)
        only_stats : bool, optional
            Specifies if only file metadata is loaded (True), or if the full article content is loaded (False). On average,
            only_stats loads about 1/10 and is recommended for quick metadata-based analysis. (Default is False)
        merge_body : bool, optional
            Specifies if the body field should be merged with the snippet and this last column being dropped.
            (default is False)
        Returns
        -------
        pandas.DataFrame
            A single Pandas Dataframe with the content from all read files.
        """
        format_suffix = file_format.lower()
        r_df = pd.DataFrame()
        for filename in os.listdir(folderpath):
            if filename.lower().endswith("." + format_suffix):
                t_df = self.read_file(folderpath + "/" + filename, only_stats, merge_body)
                r_df = pd.concat([r_df, t_df])
        return r_df



"""Taxonomy class implementation."""
from io import StringIO

import pandas as pd
from factiva.core import UserKey, const, req
from factiva.core.tools import validate_type


class Taxonomy():
    """Class that represents the taxonomy available within the Snapshots API.

    Parameters
    ----------
    user_key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    user_stats : boolean, optional (Default: False)
        Indicates if user statistics have to be pulled from the API. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.

    Examples
    --------
    Creating a taxonomy instance providing the user key
        >>> t = Taxonomy(u='abcd1234abcd1234abcd1234abcd1234')

    Creating a taxonomy instance with an existing UserKey instance
        >>> u = UserKey('abcd1234abcd1234abcd1234abcd1234')
        >>> t = Taxonomy(user_key=u)

    """

    categories = []

    def __init__(self, user_key=None, user_stats=False):
        """Class initializer."""
        self.user_key = UserKey.create_user_key(user_key, user_stats)
        self.categories = self.get_categories()

    def get_categories(self):
        """Request for a list of available taxonomy categories.

        Returns
        -------
        List of available taxonomy categories.

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        This method is called with in the __init__ method, so the categories can be accessed as is.
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.categories)
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        Calling the method on its own
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.get_categories())
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        """
        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_TAXONOMY_BASEPATH}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            return [entry['attributes']['name'] for entry in response.json()['data']]

        raise RuntimeError('API Request returned an unexpected HTTP status')

    def get_category_codes(self, category):
        """Request for available codes in the taxonomy for the specified category.

        Parameters
        ----------
        category : str
            String with the name of the taxonomy category to request the codes from

        Returns
        -------
        Dataframe containing the codes for the specified category

        Raises
        ------
        ValueError: When category is not of a valid type
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Getting the codes for the 'industries' category
            >>> taxonomy = Taxonomy()
            >>> industry_codes = taxonomy.get_category_codes('industries')
            >>> print(industry_codes)
                    code                description
            0     i25121             Petrochemicals
            1     i14001         Petroleum Refining
            2       i257            Pharmaceuticals
            3     iphrws  Pharmaceuticals Wholesale
            4       i643     Pharmacies/Drug Stores

        """
        validate_type(category, str, 'Unexpected value: category value must be string')

        response_format = 'csv'

        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_TAXONOMY_BASEPATH}/{category}/{response_format}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            return pd.read_csv(StringIO(response.content.decode()))

        raise RuntimeError('API Request returned an unexpected HTTP Status')

    def get_single_company(self, code_type, company_code):
        """Request information about a single company.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        company_code : str
            String containing the company code

        Returns
        -------
        DataFrame containing the company information

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Get the company data using the code type 'isin' and the company code 'ABCNMST00394'
            >>> taxonomy = Taxonomy()
            >>> company_data = taxonomy.get_single_company('isin', 'ABCNMST00394')
            >>> print(company_data)
                         id  fcode           common_name
            0  ABCNMST00394  ABCYT  Systemy Company S.A.

        """
        validate_type(code_type, str, 'Unexpected value: code_type must be str')
        validate_type(company_code, str, 'Unexpected value: company must be str')

        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_COMPANIES_BASEPATH}/{code_type}/{company_code}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            response_data = response.json()
            return pd.DataFrame.from_records([response_data['data']['attributes']])

        raise RuntimeError('API Request returned an unexpected HTTP status')

    def get_multiple_companies(self, code_type, company_codes):
        """Request information about a list of companies.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        companies_codes : list
            List containing the company codes to request information about

        Returns
        -------
        DataFrame containing the company information

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Get multiple companies data using the code type 'isin' and a company codes list
            >>> taxonomy = Taxonomy()
            >>> companies_data = taxonomy.get_multiple_companies('isin', ['ABC3E53433100', 'XYZ233341067', 'MN943181045'])
            >>> print(companies_data)
                         id   fcode      common_name
            0  ABC3E5343310  MCABST  M**************
            1  XYZ233341067   AXYZC    A************
            2  MN9431810453     MNM     M***********

        """
        validate_type(code_type, str, 'Unexpected value: code_type must be str')
        validate_type(company_codes, list, 'Unexpected value: companies must be list')
        for single_company_code in company_codes:
            validate_type(single_company_code, str, 'Unexpected value: each company in companies must be str')

        headers_dict = {
            'user-key': self.user_key.key
        }

        payload_dict = {
            "data": {
                "attributes": {
                    "ids": company_codes
                }
            }
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_COMPANIES_BASEPATH}/{code_type}'

        response = req.api_send_request(method='POST', endpoint_url=endpoint, headers=headers_dict, payload=payload_dict)

        if response.status_code == 200 or response.status_code == 207:
            response_data = response.json()
            return pd.DataFrame.from_records(response_data['data']['attributes']['successes'])
        raise RuntimeError(f'API Request returned an unexpected HTTP status with message: {response.text}')

    def get_company(self, code_type, company_code=None, companies_codes=None):
        """Request information about either a single company or a list of companies.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        company_code: str, Optional
            Single company code to request data about. Not compatible with companies_codes.
        companies_codes: List[str], Optional
            List of string that contains the company codes to request data about. Not compatible with company_code.

        Returns
        -------
        Dataframe with the information about the requested company(ies)

        Raises
        ------
        ValueError: When any given argument is not of the expected type
        RuntimeError:
            - When both company and companies arguments are set
            - When API request returns unexpected error

        Examples
        --------
        Get data for a single company using the code type 'isin'
            >>> taxonomy = Taxonomy()
            >>> single_company_data = taxonomy.get_company('isin', company_code='ABCNMST00394')
            >>> print(single_company_data)
                         id  fcode           common_name
            0  ABCNMST00394  ABCYT  Systemy Company S.A.

        Get data for multiple companies sugin the code type 'isin'
            >>> taxonomy = Taxonomy()
            >>> multiple_companies_data = taxonomy.get_company('isin', company_codes=['ABC3E53433100', 'XYZ233341067', 'MN943181045'])
            >>> print(multiple_companies_data)
                         id   fcode      common_name
            0  ABC3E5343310  MCABST  M**************
            1  XYZ233341067   AXYZC    A************
            2  MN9431810453     MNM     M***********

        """
        if company_code is not None and companies_codes is not None:
            raise RuntimeError('company and companies parameters cannot be set simultaneously')

        if company_code is not None:
            return self.get_single_company(code_type, company_code)

        if companies_codes is not None:
            return self.get_multiple_companies(code_type, companies_codes)

        return None



"""Implement query class definition."""
from factiva.core import const
from factiva.core.tools import validate_type
from factiva.news.bulknews import BulkNewsQuery


def validate_field_options(field, available_options):
    """Validate that field is among the available options.

    Parameters
    ----------
    field:
        field to be validated, could be any type.
    available_options:
        options amongst field should be located.

    """
    if field not in available_options:
        raise ValueError(f'Value of file_format is not within the allowed options: {available_options}')


class SnapshotQuery(BulkNewsQuery):
    """Implement Snapshot Query class definition."""
    limit = 0
    file_format = ''
    frequency = ''
    date_field = ''
    group_by_source_code = False
    top = 0

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        where,
        includes=None,
        excludes=None,
        select_fields=None,
        limit=0,
        file_format=const.API_AVRO_FORMAT,
        frequency=const.API_MONTH_PERIOD,
        date_field=const.API_PUBLICATION_DATETIME_FIELD,
        group_by_source_code=False,
        top=10
    ):
        """Initialize query class instance."""
        super().__init__(where=where, includes=includes, excludes=excludes, select_fields=select_fields)

        validate_type(limit, int, "Unexpected value for limit")
        if limit >= 0:
            self.limit = limit
        else:
            raise ValueError("Limit value is not valid or not positive")

        validate_type(file_format, str, "Unexpected value for file_format")
        file_format = file_format.lower().strip()
        validate_field_options(file_format, const.API_EXTRACTION_FILE_FORMATS)
        self.file_format = file_format

        validate_type(frequency, str, "Unexpected value for frequency")
        frequency = frequency.upper().strip()
        validate_field_options(frequency, const.API_DATETIME_PERIODS)
        self.frequency = frequency

        validate_type(date_field, str, "Unexpected value for date_field")
        date_field = date_field.lower().strip()
        validate_field_options(date_field, const.API_DATETIME_FIELDS)
        self.date_field = date_field

        validate_type(group_by_source_code, bool, "Unexpected value for group_by_source_code")
        self.group_by_source_code = group_by_source_code

        validate_type(top, int, "Unexpected value for top")
        if top >= 0:
            self.top = top
        else:
            raise ValueError('Top value is not valid')

    def get_explain_query(self):
        """Obtain Base Query."""
        query_dict = self.get_base_query()
        return query_dict

    def get_analytics_query(self):
        """Obtain analytics Query."""
        query_dict = self.get_base_query()
        query_dict["query"].update({"frequency": self.frequency})
        query_dict["query"].update({"date_field": self.date_field})
        query_dict["query"].update({"group_by_source_code": self.group_by_source_code})
        query_dict["query"].update({"top": self.top})
        return query_dict

    def get_extraction_query(self):
        """Obtain the string querying Factiva."""
        query_dict = self.get_base_query()

        if self.limit > 0:
            query_dict["query"].update({"limit": self.limit})

        query_dict["query"].update({"format": self.file_format})

        return query_dict

    def __repr__(self):
        """Create string representation for Query Class."""
        return self.__str__()

    def __str__(self, detailed=False, prefix='  |-', root_prefix=''):
        """Create string representation for Query Class."""
        pprop = self.__dict__.copy()

        ret_val = f'{root_prefix}{str(self.__class__)}\n'
        if detailed:
            ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        else:
            ret_val += f'{prefix}where: '
            ret_val += (self.where[:77] + '...') if len(self.where) > 80 else self.where
            ret_val += f'\n{prefix}...'
            del pprop['where']
            ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        return ret_val



"""Implement actions with Bulk news such as Snapshot and Stream."""
import time
import os

from pathlib import Path
from datetime import datetime

from factiva.core import const, UserKey
from factiva.core.tools import mask_string
from factiva.core import req


def parse_field(field, field_name):
    """Parse field according to field type.

    Parameters
    ----------
    field: str, dict
        field to be parsed. When a dictionary is given, it will return it
        as is. When a string is provided it will return the eval version
        of it, in order to return a dict
    field_name: str
        name of the field to be parsed. It is displayed in the error message
        when the field type is not valid.

    """
    if isinstance(field, dict):
        return field

    if isinstance(field, str):
        return eval(field)

    raise ValueError(f'Unexpected value for {field_name}')


class BulkNewsBase():
    """Represent actions with Bulk news such as Snapshot and Stream.

    Class that serves as a base for the diferent classes representing actions with bulk news such as Snapshot and Stream
    This class is not to be instantiated directly.

    Parameters
    ----------
    key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the constructor will try to obtain its
        value from the FACTIVA_USERKEY environment variable.
    stats : boolean, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation fills account detail properties
        along with maximum, used and remaining values. It may take several seconds to complete.

    """

    user_key = None

    _request_datetime = ''
    _job_status = ''

    def __init__(
        self,
        user_key=None,
        user_stats=False
    ):
        """Initialize class constructor."""
        self.user_key = UserKey.create_user_key(user_key, user_stats)

    def load_data(self):
        """Load generic data."""
        pass

    def process_operation(self):
        """Process generic operation."""
        pass

    def __str__(self):
        """Create string representation for BulkNewsBase Class."""
        pprop = self.__dict__.copy()
        del pprop['user_key']
        masked_key = mask_string(self.user_key.user_key)
        user_class = str(self.user_key.__class__)

        ret_val = str(self.__class__) + '\n'
        ret_val += f'  user_key = {masked_key} ({user_class})\n'
        ret_val += '  '.join(('{} = {}\n'.format(item, pprop[item]) for item in pprop))
        return ret_val

    def __repr__(self):
        """Create string representation for BulkNewsBase Class."""
        return self.__str__()


class BulkNewsQuery():  # pylint: disable=too-few-public-methods
    """Define base class for queries to make API requests.

    Class that serves as a base for the queries used to make requests to the APIs.

    Parameters
    ----------
    where: str
        String containing the query where clause
    includes: str or dict, Optional (Default: None)
        Contains the query includes clause, which defines the codes to include for specific fields.
        - When a dict is given it is assigned as it is.
        - When a string is given, it is expected to have a JSON format and is parsed to be stored as a dict
    excludes: str or dict, Optional (Default: None)
        Contains the query excludes clause, which defines the codes to exclude for especific fields.
        - When a dict is given it is assigned as it is.
        - When a string is given, it is expected to have a JSON format and is parsed to be stored as a dict.
    select_fields: str or list, Optional (Default: None)
        Contains the query select clause, which defines the fields to be returned by the query.
        - When a list is given it is assigned as it is
        - When a string is given, it is expected to have a JSON array format and is parsed to be stored as a list

    """

    where = ""
    includes = ""
    excludes = ""
    select_fields = ""

    def __init__(
        self,
        where,
        includes=None,
        excludes=None,
        select_fields=None
    ):
        """Initialize class."""
        if isinstance(where, str):
            self.where = where  # TODO: Validate syntax if possible. At least it's a string.
        else:
            raise ValueError("Unexpected where clause.")

        if includes:
            self.includes = parse_field(includes, 'includes')

        if excludes:
            self.excludes = parse_field(excludes, 'excludes')

        if select_fields:
            if isinstance(select_fields, list):
                self.select_fields = select_fields  # TODO: Validate syntax if possible
            elif isinstance(select_fields, str):
                self.select_fields = eval(select_fields)
            else:
                raise ValueError("Unexpected value for select_fields")

    def get_base_query(self):
        """Create the basic query to be used within the Factiva Snapshots API and the Factiva Streams API.

        Returns
        -------
        Dictionary containing the fields that where assigned to the query.

        """
        query_dict = {
            "query": {
                "where": self.where
            }
        }

        if self.select_fields:
            query_dict["query"].update({"select": self.select_fields})

        if self.includes:
            query_dict["query"].update({"includes": self.includes})

        if self.excludes:
            query_dict["query"].update({'excludes': self.excludes})

        return query_dict


class BulkNewsJob():
    """Represent the operations of the base class.

    Base class to represent the operations that can be done using Factiva Snapshots API or Factiva Streams API.

    The operations are done in two steps:
    1. Submit job:
        This will make a request to the corresponding API to start a new job and, if successful, the API will return a job_id that can be used to monitor the status of the job
    2. Get job results:
        This will make a request to the API to get the status of the job. If the job is finished, the results are saved.

    This class is not to be instantiated directly, rather to extend in order to implement the missing methods.

    Parameters
    ----------
    user_key:
        user_key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    user_key_stats : boolean, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.

    """

    job_id = ''
    job_state = ''
    submitted_datetime = 0
    link = ''
    files = []

    def __init__(self, user_key=None, user_key_stats=False):
        """Initialize Bulk news job class."""
        self.job_id = ''
        self.job_state = ''
        self.submitted_datetime = datetime.now()
        self.link = ''
        self.user_key = UserKey.create_user_key(user_key, user_key_stats)

    def get_endpoint_url(self) -> str:
        """Create the URL for the API endpoint to send the request to sumbit a job, according to the kind of job that is being created.

        This method needs to be defined by each class that extends this one.

        Returns
        -------
        String containing the endpoint URL to sumbit the job to.

        Raises
        ------
        - NotImplementedError when the method has not been defined by the extending class

        """
        raise NotImplementedError('Method has not been defined')

    def get_job_id(self, source: dict) -> str:
        """Obtain the job_id from the source parameter. The job_id is defined diferently according to the type of job.

        Parameters
        ----------
        source: dict
            Dictionary containing the data from where to obtain the id from. Usually is a dictionary representation of the response given by the API

        Returns
        -------
        String containing the job id

        Raises
        ------
        - NotImplementedError when the method has not been defined by the extending class

        """
        raise NotImplementedError('Method has not been defined')

    def set_job_data(self, source: dict):
        """Obtain the data for the job, based on the response from the API.

        Parameters
        ----------
        source: dict
            Dictionary containing the data from where to obtain the job results from. Usually is a dictionary representation of the response given by the API

        Returns
        -------
        No return value

        Raises
        ------
        - NotImplementedError when the method has not been defined by the extending class

        """
        raise NotImplementedError('Method has not been defined')

    def submit_job(self, payload=None) -> bool:
        """Submit a new job to be processed to the Factiva Snapshots API or Streams API.

        Submits a new job to be processed to the Factiva Snapshots API or Streams API.
        On a successful response from the API, saves the link of the job as well as the job_id on the caller instance.

        Parameters
        ----------
        payload: dict or str, Optional
            Contains the payload required to create the new job. Usually some, the query to be sent to the API.
            - When a dict is given, it is converted into a string
            - When a string is given, it is used as it is.

        Returns
        -------
        Boolean: True if the data was retrieved successfully. An Exception
            otherwise.

        Raises
        ------
        - RuntimeError when the API response does not have a 201 status code.

        """
        self.submitted_datetime = datetime.now()

        headers_dict = {
                'user-key': self.user_key.key,
                'Content-Type': 'application/json'
            }
        response = req.api_send_request(method='POST', endpoint_url=self.get_endpoint_url(), headers=headers_dict, payload=payload)

        if response.status_code == 201:
            response_data = response.json()
            self.job_id = self.get_job_id(response_data)
            self.job_state = response_data['data']['attributes']['current_state']
            self.link = response_data['links']['self']
        elif response.status_code == 400:
            raise ValueError(f'Invalid Query [{response.text}]')
        else:
            raise RuntimeError(f'API request returned an unexpected HTTP status, with content [{response.text}]')
        return True

    def get_job_results(self) -> bool:
        """Make a request to the API using the link of the job to get its status.

        Makes a request to the API using the link of the job to get its status. If the job has been completed,
        obtains the results of the job.

        Returns
        -------
        Boolean : True if the submission was successful. An Exception otherwise.

        Raises
        ------
        - RuntimeError when the job does not have a link to request the status to or when the response from the API is 404, meaning that the job was not found.
        Also, when the API returns a status code different than 200.

        """
        if self.link == '':
            raise RuntimeError('Job has not yet been submitted or Job ID was not set')

        headers_dict = {
            'user-key': self.user_key.key,
            'Content-Type': 'application/json'
        }

        response = req.api_send_request(method='GET', endpoint_url=self.link, headers=headers_dict)

        if response.status_code == 200:
            response_data = response.json()
            self.job_state = response_data['data']['attributes']['current_state']
            if self.job_state == const.API_JOB_DONE_STATE:
                self.set_job_data(response_data)
        elif response.status_code == 404:
            raise RuntimeError('Job ID does not exist.')
        else:
            raise RuntimeError(f'API request returned an unexpected HTTP status, with content [{response.text}]')
        return True

    def process_job(self, payload=None) -> bool:
        """Submit a new job to be processed, wait until the job is completed and then retrieves the job results.

        Parameters
        ----------
        payload: dict or str, Optional
            Contains the payload required to create the new job.
            - When a dict is given, it is converted into a string
            - When a string is given, it is used as it is.

        Returns
        -------
        Boolean : True if the explain processing was successful. An Exception
            otherwise.

        Raises
        ------
        - RuntimeError when the job returns and unexpected state
        - Exception when the job has failed to complete

        """
        self.submit_job(payload=payload)
        self.get_job_results()

        while self.job_state != const.API_JOB_DONE_STATE:
            if self.job_state not in const.API_JOB_EXPECTED_STATES:
                raise RuntimeError('Unexpected job state')
            if self.job_state == const.API_JOB_FAILED_STATE:
                raise Exception('Job failed')

            time.sleep(const.API_JOB_ACTIVE_WAIT_SPACING)
            self.get_job_results()

        return True

    def download_file(self, endpoint_url: str, download_path: str):
        """Download a file from a job, using the file URL and stores them in download_path.

        Parameters
        ----------
        endpoint_url: str
            String containing the URL to download the file from
        download_path: str
            String containing the path where to store the downloaded file

        Returns
        -------
        Boolean : True if the download was successful. An Exception otherwise.

        Raises
        ------
        - RuntimeException when the response from the API is not successful

        """
        headers_dict = {
                'user-key': self.user_key.key
            }
        response = re.api_send_request(method='GET', endpoint_url=endpoint_url, headers=headers_dict)

        if response.status_code == 200:
            with open(download_path, 'wb') as download_file_path:
                download_file_path.write(response.content)
        else:
            raise RuntimeError(f'API request returned an unexpected HTTP status, with content [{response.text}]')
        return True

    def download_job_files(self, download_path=None):
        """Download all the files from a job ans stores them in the given download_path.

        If no download path is given, the files are stored in a folder with the name of the job_id.

        Parameters
        ----------
        download_path: str, Optional
            String containing the path where to store the downloaded files.
            If not provided, the files are stored in a folder named after the job_id. If such folder does not exists, it is created in the current working directory.

        Returns
        -------
        Boolean : True if the files were correctly downloaded, False otherwise.

        Raises
        ------
        - RuntimeError when there are no files available for download

        """
        if download_path is None:
            download_path = os.path.join(os.getcwd(), self.job_id)
        Path(download_path).mkdir(parents=True, exist_ok=True)

        if len(self.files) > 0:
            for file_uri in self.files:
                file_name = file_uri.split('/')[-1]
                local_path = f'{download_path}/{file_name}'
                self.download_file(file_uri, local_path)
        else:
            raise RuntimeError('No files available for download')
        return True

    def __repr__(self):
        """Create string representation for BulkNews Class."""
        return self.__str__()

    def __str__(self, detailed=True, prefix='  |-', root_prefix=''):
        """Create string representation for BulkNews Class."""
        pprop = self.__dict__.copy()
        ret_val = str(self.__class__)
        ret_val += '\n'
        if self.job_id == '':
            ret_val += f'{prefix}<Empty>'
        else:
            ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        return ret_val



from setuptools import setup

with open("README.rst", "r") as fh:
    long_desc = fh.read()

with open('src/factiva/news/__version__.py') as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            continue

setup(
    name='factiva-news',
    version=version,
    description='Python package to interact with Factiva news-related APIs. Services are described in the Dow Jones Developer Platform.',
    long_description=long_desc,
    long_description_content_type='text/x-rst',
    license='MIT',
    author='Dos Jones Customer Engineers',
    author_email='customer.solutions@dowjones.com',
    # Warning: the folder 'factiva' should NOT have an __init__.py file to avoid conflicts with the same namespace across other packages
    package_dir={'': 'src'},
    packages=['factiva.news', 'factiva.news.snapshot', 'factiva.news.stream', 'factiva.news.taxonomy'],
    url='https://developer.dowjones.com/',
    project_urls={
            "GitHub": "https://github.com/dowjones/factiva-news-python",
            "Documentation": "https://factiva-news-python.readthedocs.io/",
            "Bug Tracker": "https://github.com/dowjones/factiva-news-python/issues",
        },

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   1 - Planning
        #   2 - Pre-Alpha
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Financial and Insurance Industry',
        'Operating System :: OS Independent',
        'Topic :: Office/Business :: News/Diary',
        'Topic :: Office/Business :: Financial :: Investment',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='news, news aggregator, risk, compliance, nlp, alternative data',
    python_requires='>=3.6',
    install_requires=['factiva-core>=0.1.11', 'fastavro', 'google-cloud-core', 'google-cloud-pubsub']
)



"""
    Define methods and properties for Snapshot class.
"""
__all__ = ['Snapshot', 'SnapshotQuery', 'SnapshotFiles',
    'AnalyticsJob', 'ExplainJob', 'ExtractionJob', 'UpdateJob',
    'query', 'jobs', 'snapshot', 'files']

from .query import SnapshotQuery
from .jobs import AnalyticsJob, ExplainJob, ExtractionJob, UpdateJob
from .snapshot import Snapshot
from .files import SnapshotFiles



"""Implement different helper functions and classes for jobs."""
import pandas as pd


from factiva.core import const
from factiva.news.bulknews import BulkNewsJob


class ExplainJob(BulkNewsJob):
    """Represent the operation of creating an explain from Factiva Snapshots API."""

    document_volume = 0

    def __init__(self, user_key):
        """Initialize class."""
        super().__init__(user_key=user_key)
        self.document_volume = 0

    # It's important for these methods to be class methods since they
    # are overrriden methods from the base class.
    # pylint: disable=no-self-use
    def get_endpoint_url(self):
        """Get endpoint URL."""
        return f'{const.API_HOST}{const.API_SNAPSHOTS_BASEPATH}{const.API_EXPLAIN_SUFFIX}'

    # pylint: disable=no-self-use
    def get_job_id(self, source):
        """Get job ID."""
        return source['data']['id']

    def set_job_data(self, source):
        """Set job data."""
        self.document_volume = source['data']['attributes']['counts']


class AnalyticsJob(BulkNewsJob):
    """Represent the operation of creating Analtyics from Factiva Snapshots API."""
    data = []

    def __init__(self, user_key):
        """Initialize class."""
        super().__init__(user_key=user_key)
        self.data = []

    # pylint: disable=no-self-use
    def get_endpoint_url(self):
        """Get endpoint URL."""
        return f'{const.API_HOST}{const.API_ANALYTICS_BASEPATH}'

    # pylint: disable=no-self-use
    def get_job_id(self, source):
        """Get job ID."""
        return source['data']['id']

    def set_job_data(self, source):
        """Set job data."""
        self.data = pd.DataFrame(source['data']['attributes']['results'])


class ExtractionJob(BulkNewsJob):
    """Class that represents the operation of creating a Snapshot from Factiva Snapshots API."""
    files = []
    file_format = ''

    def __init__(self, snapshot_id=None, user_key=None):
        """Initialize class."""
        super().__init__(user_key=user_key)
        self.files = []
        self.file_format = ''

        if snapshot_id and user_key:
            self.job_id = snapshot_id
            self.link = f'{const.API_HOST}{const.API_SNAPSHOTS_BASEPATH}/dj-synhub-extraction-{self.user_key.key}-{snapshot_id}'

    # pylint: disable=no-self-use
    def get_endpoint_url(self):
        """Obtain endpoint URL."""
        return f'{const.API_HOST}{const.API_SNAPSHOTS_BASEPATH}'

    # pylint: disable=no-self-use
    def get_job_id(self, source):
        """Obtain Job ID."""
        return source['data']['id'].split('-')[-1]

    def set_job_data(self, source):
        """Set job data."""
        self.file_format = source['data']['attributes']['format']
        file_list = source['data']['attributes']['files']
        self.files = [file_item['uri'] for file_item in file_list]

    def process_job(self, payload=None, path=None):
        """Override method from parent class to call the method for downloading the files once the snapshot has been completed.

        Overrides method from parent class to call the method for downloading the files once the snapshot has been completed.

        Parameters
        ----------
        payload: str, Optional
            String containg the snapshot instance.

        path: str, Optional
            String containg the path where to store the snapshots files that are downloaded from the snapshot.
            If no path is given, the files will be stored in a folder named after the snapshot_id in the current working directory.

        """
        super().process_job(payload)
        self.download_job_files(path)


class UpdateJob(ExtractionJob):
    """Represent the Snapshot Updates.

    Class that represents the Snapshot Updates. There can be three types of updates: additions, replacements and deletes.

    Parameters
    ----------
    update_type: str, Optional
        String describing the type of update that this job represents. Requires snapshot_id to be provided as well. Not compatible with update_id
    snapshot_id: str, Optional
        String containing the id of the snapshot that is being updated. Requires update_type to be provided as well. Not compatible with update_id
    update_id: str, Optional
        String containing the id of an update job that has been created previously. Both update_type and snapshot_id can be obtained from this value.
        Not compatible with update_type nor snapshot_id

    Raises
    ------
    - Exception when fields that are not compatible are provided or when not enough parameters are provided to create the job.

    """
    update_type = None
    snapshot_id = None

    def __init__(self, update_type=None, snapshot_id=None, update_id=None, user_key=None):
        """Construct class instance."""
        super().__init__(user_key=user_key)

        if update_id and (update_type or snapshot_id):
            raise Exception('update_id parameter is not compatible with update_type and snapshot id')

        if update_id:
            self.job_id = update_id
            self.update_type = update_id.split('-')[1]
            self.snapshot_id = update_id.split('-')[0]
            self.link = f'{const.API_HOST}{const.API_SNAPSHOTS_BASEPATH}/dj-synhub-extraction-{self.user_key.key}-{update_id}'
            self.get_job_results()

        elif update_type and snapshot_id:
            self.update_type = update_type
            self.snapshot_id = snapshot_id
        else:
            raise Exception('Not enough parameters to create an update job')

    def get_endpoint_url(self):
        """Get endpoint URL."""
        return f'{const.API_HOST}{const.API_EXTRACTIONS_BASEPATH}/dj-synhub-extraction-{self.user_key.user_key}-{self.snapshot_id}/{self.update_type}'

    def get_job_id(self, source):
        """Get job ID from source."""
        # UPDATE_ID FORMAT: {API_URL}/dj-synhub-extraction-{USER-KEY}-{SNAPSHOT_ID}-{UPDATE_TYPE}-{DATETIME}
        return '-'.join(source['data']['id'].split('-')[-3:])
